# PYXPLOT.PY
# Dominic Ford
# 26/08/2006

exitting = 0

import gp_children
import gp_settings
import gp_eval
import gp_fit
import gp_plot
import gp_text
import gp_help
import gp_version
import gp_spline
from gp_autocomplete import *
from gp_error import *

import os
import sys
import glob
import signal
import readline
from math import *
import re

# FLOATPRINT(): Prints a float in floating point or exponential format, as appropriate

def floatprint(x):
  if ((fabs(x) < 1e10) and (fabs(x) > 1e-6)): return "%f"%x
  else:                                       return "%e"%x

# READ_COORD_SYST(): Read screen, graph, first or second modifiers
# Returns modifier and string of unprocessed text

def read_coord_syst(string):
  if  ((string == None) or (len(string) == 0) or
       autocomplete(string, "first", 1)          ): return ['first','']
  elif autocomplete(string, "second", 1)          : return ['second','']
  elif autocomplete(string, "graph", 1)           : return ['graph','']
  elif autocomplete(string, "screen", 1)          : return ['screen','']
  else:
   test = re.match(r"axis(\d\d*)$",string)
   if (test != None): return [string,'']
   return ['first',string]

# Directive Show

def directive_show(line,linelist):
  wordlist = linelist[1:]
  if (len(wordlist) == 0):
    gp_error(gp_text.show)
  else:
    for word in wordlist:
      if autocomplete(word, "all",1):
        directive_show("",["cabbage","settings","axes_","linestyles","variables","functions"])
        continue
      outstring = ""
      if autocomplete(word, "settings", 1) or autocomplete(word, "origin", 1):    outstring += "Plot Offset:  (%f,%f)\n"%(gp_settings.settings['ORIGINX'],gp_settings.settings['ORIGINY'])
      if autocomplete(word, "settings", 1) or autocomplete(word, "multiplot", 1): outstring += "Multiplot:     %s\n"%gp_settings.settings['MULTIPLOT']
      if autocomplete(word, "settings", 1) or autocomplete(word, "display", 1)  : outstring += "Display:       %s\n"%gp_settings.settings['DISPLAY']
      if autocomplete(word, "settings", 1) or autocomplete(word, "title", 1):     outstring += "Plot Title:   '%s' at offset (%f,%f)\n"%(gp_settings.settings['TITLE'],gp_settings.settings['TIT_XOFF'],gp_settings.settings['TIT_YOFF'])
      if autocomplete(word, "settings", 1) or autocomplete(word, "terminal", 1):  outstring += "Terminal type: %s\n"%gp_settings.settings['TERMTYPE']
      if autocomplete(word, "settings", 1) or autocomplete(word, "output", 1):    outstring += "Output fname:  %s\n"%gp_settings.settings['OUTPUT']
      if autocomplete(word, "settings", 1) or autocomplete(word, "backup", 1):    outstring += "File backups:  %s\n"%gp_settings.settings['BACKUP']
      if autocomplete(word, "settings", 1) or autocomplete(word, "terminal", 1):  outstring += "Enhanced eps:  %s\n"%gp_settings.settings['ENHANCED']
      if autocomplete(word, "settings", 1) or autocomplete(word, "terminal", 1):  outstring += "Landscape mode:%s\n"%gp_settings.settings['LANDSCAPE']
      if autocomplete(word, "settings", 1) or autocomplete(word, "terminal", 1):  outstring += "Invert colours:%s (only relevant when output is sent to a bitmap terminal -- gif, jpg or png)\n"%gp_settings.settings['TERMINVERT']
      if autocomplete(word, "settings", 1) or autocomplete(word, "terminal", 1):  outstring += "Transparency:  %s (only relevant when output is sent to gif or png terminals, which support transparency)\n"%gp_settings.settings['TERMTRANSPARENT']
      if autocomplete(word, "settings", 1) or autocomplete(word, "dpi", 1):       outstring += "Output DPI:    %f (only relevant when output is sent to a bitmap terminal -- gif, jpg or png)\n"%gp_settings.settings['DPI']
      if(autocomplete(word, "settings", 1) or autocomplete(word, "width", 1) or
                                               autocomplete(word, "size", 1)):    outstring += "Output width:  %s\n"%floatprint(gp_settings.settings['WIDTH'])
      if autocomplete(word, "settings", 1) or autocomplete(word, "size", 1):
       if (gp_settings.settings['AUTOASPECT'] == "ON"): outstring += "Output aspect: <auto>\n"
       else                                           : outstring += "Output aspect: %f\n"%gp_settings.settings['ASPECT']
      if(autocomplete(word, "settings", 1) or autocomplete(word, "pointsize", 1) or
                                              autocomplete(word, "ps",2)       ): outstring += "Pointsize:     %f\n"%gp_settings.settings['POINTSIZE']
      if (autocomplete(word, "settings", 1) or
          autocomplete(word, "pointlinewidth",1) or autocomplete(word, "plw",3)): outstring += "Pointlinewidth:%f\n"%gp_settings.settings['POINTLINEWIDTH']
      if autocomplete(word, "settings", 1) or autocomplete(word, "bar",1):        outstring += "Bar width:     %f (the size of the strokes at the end of errorbars, set with 'set bar')\n"%gp_settings.settings['BAR']
      if autocomplete(word, "settings", 1) or autocomplete(word, "fontsize",1):   outstring += "Fontsize:      %d (-4 is smallest, 5 largest)\n"%gp_settings.settings['FONTSIZE']
      if autocomplete(word, "settings", 1) or autocomplete(word, "textcolour",1): outstring += "Text colour:   %s\n"%gp_settings.settings['TEXTCOLOUR']
      if autocomplete(word, "settings", 1) or autocomplete(word, "texthalign",1): outstring += "Text halign:   %s\n"%gp_settings.settings['TEXTHALIGN']
      if autocomplete(word, "settings", 1) or autocomplete(word, "textvalign",1): outstring += "Text valign:   %s\n"%gp_settings.settings['TEXTVALIGN']
      if autocomplete(word, "settings", 1) or autocomplete(word, "axescolour",1): outstring += "Axes colour:   %s\n"%gp_settings.settings['AXESCOLOUR']
      if autocomplete(word, "settings", 1) or autocomplete(word, "gridmajcolour",1): outstring += "Grid major col:%s\n"%gp_settings.settings['GRIDMAJCOLOUR']
      if autocomplete(word, "settings", 1) or autocomplete(word, "gridmincolour",1): outstring += "Grid minor col:%s\n"%gp_settings.settings['GRIDMINCOLOUR']
      if(autocomplete(word, "settings", 1) or autocomplete(word, "linewidth", 1) or
                                              autocomplete(word, "lw",2)       ): outstring += "Linewidth:     %f\n"%gp_settings.settings['LINEWIDTH']
      if(autocomplete(word, "settings", 1) or autocomplete(word, "data", 1) or
                                               autocomplete(word, "style",1)):    outstring += "Data style:    %s (the default plotting style for datafiles, if none is specified)\n"%gp_settings.settings['DATASTYLE']
      if(autocomplete(word, "settings", 1) or autocomplete(word, "function", 1) or
                                               autocomplete(word, "style",1)):    outstring += "Function style:%s (the default plotting style for functions, if none is specified)\n"%gp_settings.settings['FUNCSTYLE']
      if autocomplete(word, "settings", 1) or autocomplete(word, "samples",1):    outstring += "Samples        %d (no of samples used when plotting functions)\n"%gp_settings.settings['SAMPLES']
      if autocomplete(word, "settings", 1) or autocomplete(word, "colour",1):     outstring += "Colour:        %s (when off, all plots will be monochrome)\n"%gp_settings.settings['COLOUR']
      if autocomplete(word, "settings", 1) or autocomplete(word, "key",1):        outstring += "Key:           %s (selects whether a legend appears on plots)\n"%gp_settings.settings['KEY']
      if autocomplete(word, "settings", 1) or autocomplete(word, "keycolumns",1): outstring += "Key columns:   %s\n"%gp_settings.settings['KEYCOLUMNS']
      if autocomplete(word, "settings", 1) or autocomplete(word, "key",1):        outstring += "Key position:  %s, with offset (%f,%f)\n"%(gp_settings.settings['KEYPOS'],gp_settings.settings['KEY_XOFF'],gp_settings.settings['KEY_YOFF'])
      if autocomplete(word, "settings", 1) or autocomplete(word, "grid",1):
        outstring += "Grid:          %s\n"%gp_settings.settings['GRID']
        if (gp_settings.settings['GRID'] == 'ON'):
          outstring += "Grid Axes:     "
          for axis_n in gp_settings.settings['GRIDAXISX']: outstring += "x%d "%axis_n
          for axis_n in gp_settings.settings['GRIDAXISY']: outstring += "y%d "%axis_n
          outstring += "\n"
      if autocomplete(word, "settings", 1) or autocomplete(word, "boxwidth",1):   outstring += "Boxwidth:      %s (the default width of bars on barcharts and histograms; a negative value means automatic widths)\n"%floatprint(gp_settings.settings['BOXWIDTH'])
      if autocomplete(word, "settings", 1) or autocomplete(word, "boxfrom",1):    outstring += "BoxFrom:       %s (the vertical point from which the bars of barcharts and histograms emanate)\n"%floatprint(gp_settings.settings['BOXFROM'])
      if autocomplete(word, "settings", 1) or autocomplete(word, "palette",1):    outstring += "Palette:       %s\n"%gp_settings.colour_list

      if autocomplete(word, "linestyles", 1) or (word=="ls"):
       outstring += "\nLinestyles:\n"
       for i,definition in gp_settings.linestyles.iteritems():
         outstring += "  linestyle %d: %s\n"%(i,definition)

      if autocomplete(word, "arrows",1):
        outstring += "\nArrows:\n"
        for i,definition in gp_settings.arrows.iteritems():
          outstring += "  arrow %d: (%s %s,%s %s) to (%s %s,%s %s) %s\n"%(i, definition[0], floatprint(definition[1]), definition[2], floatprint(definition[3]), definition[4], floatprint(definition[5]), definition[6], floatprint(definition[7]), definition[8])

      if autocomplete(word, "labels",1):
        outstring += "\nText labels:\n"
        for i,definition in gp_settings.labels.iteritems():
          outstring += "  label %d: %s at (%s %s,%s %s)\n"%(i, definition[0], definition[1], floatprint(definition[2]), definition[3], floatprint(definition[4]))

      for [direction , direction_axes] in gp_settings.axes.iteritems():
       for [axis_number, axis] in direction_axes.iteritems():
        axisoutstring = ""
        if autocomplete(word, "axes_",1) or autocomplete(word, "%s%dlabel"%(direction,axis_number),1):
          axisoutstring += "  Label:     %s\n"%axis['LABEL']
        if autocomplete(word, "axes_",1) or autocomplete(word, "%s%drange"%(direction,axis_number),1) or autocomplete(word, "autoscale",1):
          if (axis['MIN'] == None): mintxt = "<auto>"
          else                    : mintxt = floatprint(axis['MIN'])
          if (axis['MAX'] == None): maxtxt = "<auto>"
          else                    : maxtxt = floatprint(axis['MAX'])
          axisoutstring += "  Range:    (%s --> %s)\n"%(mintxt,maxtxt)
        if autocomplete(word, "axes_",1) or autocomplete(word, "logscale",1):
          axisoutstring += "  Log:       %s\n"%axis['LOG']
        if autocomplete(word, "axes_",1) or autocomplete(word, "%s%dticdir"%(direction,axis_number),1):
          axisoutstring += "  Tick Direction: %s\n"%axis['TICDIR']
        if (len(axisoutstring) > 0): outstring += "\nSettings for %s%s axis:\n"%(direction,axis_number)+axisoutstring

      if autocomplete(word, "variables",1) or (word=="vars"):
        outstring += "\nVariables:\n"
        for x,y in gp_settings.variables.iteritems():
          outstring += "%s = %s\n"%(x,floatprint(y))
        outstring += "\n"

      if autocomplete(word, "functions",1) or (word=="funcs"):
        outstring += "\nUser-Defined Functions:\n"
        for x,y in gp_settings.functions.iteritems():
         for definition in y[1]:
          string = x+"("
          if (y[0] < 0):   # This is a spline
            outstring += string + "x) = spline fit to file %s.\n"%definition[0]
          else:              # This is a regular function
            ranges = " for "
            for i in range(y[0]):
             if (i != 0): string += ","
             string += definition[0][i]
             if   (i != 0) and (i != y[0]-1): ranges += ", "
             elif (i != 0) and (i == y[0]-1): ranges += " and "
             if (definition[1][i] == [None,None]): ranges += "all "+definition[0][i]
             else:
              ranges += "("
              if (definition[1][i][0] == None): ranges += "-inf"
              else                            : ranges += definition[1][i][0]
              ranges += " < "+definition[0][i]+" < "
              if (definition[1][i][1] == None): ranges += "inf"
              else                            : ranges += definition[1][i][1]
              ranges += ")"
            outstring += string+") = "+definition[2]+"\n"
            outstring += ranges+"\n"

      if (len(outstring) == 0):
        gp_error("Error: show command passed unrecognised word '%s'."%word)
        gp_error(gp_text.show)
      else:
        gp_report(outstring)
  return

# Directive Set / Unset

def directive_set_unset(command,line,linelist):
    if (len(linelist) < 2):
      if (command == "set"):   gp_error(gp_text.set)
      if (command == "unset"): gp_error(gp_text.unset)
      return

    if ((command == "set") and autocomplete(linelist[1],"title",2)): # set title
      errorstr = "Title should be placed in quotes, and followed optionally by xoffset,yoffset. Defaulting to blank title."
      gp_settings.settings['TIT_XOFF'] = gp_settings.settings['TIT_YOFF'] = 0.0 # Reset offset
      test = re.match(r"""\S*\s*\S*\s*(('|").*)""",line)
      if (test == None):
        gp_error(errorstr) ; gp_settings.settings['TITLE']="" ; return
      [title, aftertitle] = gp_eval.gp_getquotedstring(test.group(1))
      if (title == None):
        gp_error(errorstr) ; gp_settings.settings['TITLE']="" ; return
      test = re.match(r"""\s*([^\s,]*)\s*(,\s*([^\s,]*))?""",aftertitle)
      if (test == None):
        gp_error(errorstr) ; gp_settings.settings['TITLE']="" ; return
      gp_settings.settings['TITLE']=title
      try:
       if  (test.group(1) != "")                             : gp_settings.settings['TIT_XOFF'] = gp_eval.gp_eval(test.group(1),gp_settings.variables,gp_settings.functions)
       if ((test.group(3) != "") and (test.group(2) != None)): gp_settings.settings['TIT_YOFF'] = gp_eval.gp_eval(test.group(3),gp_settings.variables,gp_settings.functions)
      except KeyboardInterrupt: raise
      except:
       gp_error("Could not read positional offsets of title (%s,%s)"%(test.group(1),test.group(3)))
      return

    if (((command == "set") and autocomplete(linelist[1],"notitle",3)) or # set notitle
        ((command == "unset") and autocomplete(linelist[1],"title",2))  ):
     if (len(linelist) > 2):
      gp_error('Trailing word after "title" not expected.')
     gp_settings.settings['TITLE']    = gp_settings.settings_default['TITLE']
     gp_settings.settings['TIT_XOFF'] = gp_settings.settings_default['TIT_XOFF']
     gp_settings.settings['TIT_YOFF'] = gp_settings.settings_default['TIT_YOFF']
     return

    if ((command == "set") and (autocomplete(linelist[1],"width",1)  # set width # EXTENDED API
        or (autocomplete(linelist[1],"size",1) and (len(linelist) == 3) and (linelist[2].strip()[0] < 'A')))): # take care not to catch set size ratio
      if autocomplete(linelist[1],"size",1):
       gp_warning("Warning: 'set size <n>' is depreciated API. Reverting to pyxplot extended API function 'set width', setting width in centimetres.")
      test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
      if (test == None):
        gp_error("Error reading width.")
      else:
        try:
          width = gp_eval.gp_eval(test.group(1),gp_settings.variables,gp_settings.functions)
          if (width <= 0.0):
           gp_warning("Warning: trying to set width to %s cm.... perhaps that's a tad small?"%floatprint(width))
          else:
           gp_settings.settings['WIDTH'] = width
        except KeyboardInterrupt: raise
        except:
          gp_error("Could not read width. Should be a floating point value.")
      return

    if ((command == "unset") and autocomplete(linelist[1],"width",1)): # unset width # EXTENDED API
     if (len(linelist) > 2):
      gp_error('Trailing word after "width" not expected.')
     gp_settings.settings['WIDTH'] = gp_settings.settings_default['WIDTH']
     return

    if ((command == "set") and autocomplete(linelist[1],"boxwidth",1)):  # set boxwidth
     test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
     if (test == None):
      gp_error("Error reading boxwidth.")
     else:
      try:
       gp_settings.settings['BOXWIDTH'] = gp_eval.gp_eval(test.group(1),gp_settings.variables,gp_settings.functions)
      except KeyboardInterrupt: raise
      except:
       gp_error("Could not read boxwidth. Should be a floating point value.")
     return

    if ((command == "unset") and autocomplete(linelist[1],"boxwidth",1)): # unset boxwidth
     if (len(linelist) > 2):
      gp_error('Trailing word after "boxwidth" not expected.')
     gp_settings.settings['BOXWIDTH'] = gp_settings.settings_default['BOXWIDTH']
     return

    if ((command == "set") and autocomplete(linelist[1],"bar",2)):  # set bar
     test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
     if (test == None):
      gp_error("Error reading bar width.")
     else:
      try:
       if   (test.group(1) == "large"): gp_settings.settings['BAR'] = 1.0
       elif (test.group(1) == "small"): gp_settings.settings['BAR'] = 0.0
       else                           : gp_settings.settings['BAR'] = gp_eval.gp_eval(test.group(1),gp_settings.variables,gp_settings.functions)
      except KeyboardInterrupt: raise
      except:
       gp_error("Could not read bar width. Should be a floating point value, 'large', or 'small'.")
     return

    if ((command == "unset") and autocomplete(linelist[1],"bar",2)): # unset bar
     if (len(linelist) > 2):
      gp_error('Trailing word after "bar" not expected.')
     gp_settings.settings['BAR'] = gp_settings.settings_default['BAR']
     return

    if ((command == "set") and autocomplete(linelist[1],"palette",1)):  # set palette
     error_string = "'set palette' statement should be followed by a comma-separated list of colours to put in new palette."
     test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
     if (test == None):
      gp_error("Error reading palette.\n%s"%error_string)
     else:
      colours_new = []
      try:
       colours_new = []
       colours_in  = test.group(1).split(',')
       for colour in colours_in:
        if (colour.strip().capitalize() in gp_settings.colours): colours_new.append(colour.strip().capitalize())
        else                                                   : gp_error("Unrecognised colour '%s' in palette; skipping."%colour.strip())
       if (len(colours_new) == 0):
        gp_error(error_string)
       else:
        gp_settings.colour_list = colours_new
      except KeyboardInterrupt: raise
      except:
       gp_error("Internal error whilst reading set palette statment.\n%s"%error_string)
       gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")
     return

    if ((command == "unset") and autocomplete(linelist[1],"palette",1)): # unset palette
     if (len(linelist) > 2):
      gp_error('Trailing word after "palette" not expected.')
     gp_settings.colour_list = gp_settings.colour_list_default
     return

    if ((command == "set") and autocomplete(linelist[1],"boxfrom",4)):  # set boxfrom
     test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
     if (test == None):
      gp_error("Error reading boxfrom.")
     else:
      try:
       gp_settings.settings['BOXFROM'] = gp_eval.gp_eval(test.group(1),gp_settings.variables,gp_settings.functions)
      except KeyboardInterrupt: raise
      except:
       gp_error("Could not read boxfrom. Should be a floating point value")
     return

    if ((command == "unset") and autocomplete(linelist[1],"boxfrom",4)): # unset boxfrom
     if (len(linelist) > 2):
      gp_error('Trailing word after "boxfrom" not expected.')
     gp_settings.settings['BOXFROM'] = gp_settings.settings_default['BOXFROM']
     return

    if ((command == "set") and autocomplete(linelist[1],"size",2)): # set size
     if (len(linelist) > 2):
      if autocomplete(linelist[2],"ratio",1): # set size ratio <n>
       try:
        test = re.match(r"""^\S*\s*\S*\s*\S*\s*(.*)""",line)
        new_aspect = gp_eval.gp_eval(test.group(1),gp_settings.variables,gp_settings.functions)
        if (new_aspect <= 0.0):
          gp_error("Error: aspect ratio must be a value greater than zero")
          raise ValueError
        gp_settings.settings['ASPECT'] = new_aspect
        gp_settings.settings['AUTOASPECT'] = 'OFF'
       except KeyboardInterrupt: raise
       except:
        gp_error("Error reading aspect ratio")
       return
     if (len(linelist) == 3):
      if autocomplete(linelist[2],"noratio",1): # set size noratio
       gp_settings.settings['AUTOASPECT'] = 'ON'
       return
     if (len(linelist) == 3):
      if autocomplete(linelist[2],"square",1): # set size square
       gp_settings.settings['ASPECT'] = 1.0
       gp_settings.settings['AUTOASPECT'] = 'OFF'
       return

    if ((command == "unset") and autocomplete(linelist[1],"size",2)): # unset size
     if (len(linelist) > 2):
      gp_error('Trailing word after "size" not expected.')
     gp_settings.settings['WIDTH'] = gp_settings.settings_default['WIDTH']
     gp_settings.settings['AUTOASPECT'] = gp_settings.settings_default['AUTOASPECT']
     gp_settings.settings['ASPECT'] = gp_settings.settings_default['ASPECT']
     return

    if ((command == "set") and autocomplete(linelist[1],"multiplot",1)):  # set multiplot
     if (len(linelist) > 2):
      gp_error('Trailing word after "multiplot" not expected.')
     gp_settings.settings['MULTIPLOT'] = 'ON'
     gp_plot.multiplot_plotdesc  = [] # Wipe the plotting canvas
     gp_plot.multiplot_plotorder = []
     gp_plot.multiplot_text      = []
     gp_plot.multiplot_axes      = []
     return

    if (((command == "set") and autocomplete(linelist[1],"nomultiplot",3)) or  # set nomultiplot
        ((command == "unset") and autocomplete(linelist[1],"multiplot",1))  ):
     if (len(linelist) > 2):
      gp_error('Trailing word after "multiplot" not expected.')
     gp_settings.settings['MULTIPLOT'] = 'OFF'
     return

    if ((command == "set") and autocomplete(linelist[1],"display",1)):  # set display
     if (len(linelist) > 2):
      gp_error('Trailing word after "set display" not expected.')
     gp_settings.settings['DISPLAY'] = 'ON'
     return

    if (((command == "set") and autocomplete(linelist[1],"nodisplay",3)) or  # set nodisplay
        ((command == "unset") and autocomplete(linelist[1],"display",1))  ):
     if (len(linelist) > 2):
      gp_error('Trailing word after "set nodisplay" not expected.')
     gp_settings.settings['DISPLAY'] = 'OFF'
     return

    if ((command == "set") and autocomplete(linelist[1],"origin",2)):  # set origin
     test = re.match(r"""^\S*\s*\S*\s*([^\s,]*)(\s*,?\s*)([^\s,]*)""",line)
     if ((test == None) or (len(test.group(2)) == 0)):
      gp_error("Error reading position of origin. Use format: 'set origin x,y'.")
     else:
      try:
       gp_settings.settings['ORIGINX'] = gp_eval.gp_eval(test.group(1),gp_settings.variables,gp_settings.functions)
       gp_settings.settings['ORIGINY'] = gp_eval.gp_eval(test.group(3),gp_settings.variables,gp_settings.functions)
      except KeyboardInterrupt: raise
      except:
       gp_error("Error reading position of origin. Use format: 'set origin x,y'.")
     return

    if ((command == "unset") and autocomplete(linelist[1],"origin",2)):  # unset origin
     if (len(linelist) > 2):
      gp_error('Trailing word after "origin" not expected.')
     gp_settings.settings['ORIGINX'] = gp_settings.settings_default['ORIGINX']
     gp_settings.settings['ORIGINY'] = gp_settings.settings_default['ORIGINY']
     return

    if ((command == "set") and autocomplete(linelist[1],"samples",1)): # set samples
      test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
      if (test == None):
        gp_error("Error reading number of samples.")
      else:
        try:
          samples = int(test.group(1))
          assert samples > 0
          gp_settings.settings['SAMPLES']=samples
        except KeyboardInterrupt: raise
        except:
          gp_error("Invalid number of samples. Should be an integer value > 0.")
      return

    if ((command == "unset") and autocomplete(linelist[1],"samples",1)):  # unset samples
     if (len(linelist) > 2):
      gp_error('Trailing word after "samples" not expected.')
     gp_settings.settings['SAMPLES'] = gp_settings.settings_default['SAMPLES']
     return

    if ((command == "set") and autocomplete(linelist[1],"fontsize",1)): # set fontsize
      test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
      if (test == None):
        gp_error("Error reading fontsize.")
      else:
        try:
          value = int(test.group(1))
          if (value < -4): value = -4
          if (value >  5): value =  5
          gp_settings.settings['FONTSIZE'] = value
        except KeyboardInterrupt: raise
        except:
          gp_error("Could not read number of samples. Should be an integer value.")
      return

    if ((command == "unset") and autocomplete(linelist[1],"fontsize",1)):  # unset fontsize
     if (len(linelist) > 2):
      gp_error('Trailing word after "fontsize" not expected.')
     gp_settings.settings['FONTSIZE'] = gp_settings.settings_default['FONTSIZE']
     return

    if ((command == "set") and autocomplete(linelist[1],"textcolour",5)): # set textcolour
      test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
      if (test == None):
        gp_error("Error reading colour.")
      else:
        test2 = re.match(r"""^\d\d*$""", test.group(1))
        if (test2 != None):
         gp_settings.settings['TEXTCOLOUR'] = gp_settings.colour_list[(int(test.group(1))-1)%len(gp_settings.colour_list)]
        else:
         if (test.group(1).capitalize() in gp_settings.colours): gp_settings.settings['TEXTCOLOUR'] = test.group(1).capitalize()
         else                                                  : gp_error("Unrecognised colour '%s'."%test.group(1))
      return

    if ((command == "unset") and autocomplete(linelist[1],"textcolour",5)):  # unset textcolour
     if (len(linelist) > 2):
      gp_error('Trailing word after "textcolour" not expected.')
     gp_settings.settings['TEXTCOLOUR'] = gp_settings.settings_default['TEXTCOLOUR']
     return

    if ((command == "set") and autocomplete(linelist[1],"texthalign",5)): # set texthalign
      test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
      if (test == None):
        gp_error("Error reading text horizontal alignment.")
      else:
        errorstr = "Unrecognised horizontal alignment '%s'\nOptions are: 'left', 'centre' or 'right'.\n"%test.group(1)
        if (len(test.group(1).split()) > 1): gp_error(errorstr)
        else:
          if   autocomplete(test.group(1),"left"  ,1): gp_settings.settings['TEXTHALIGN'] = "Left"
          elif autocomplete(test.group(1),"centre",1): gp_settings.settings['TEXTHALIGN'] = "Centre"
          elif autocomplete(test.group(1),"center",1): gp_settings.settings['TEXTHALIGN'] = "Centre"
          elif autocomplete(test.group(1),"right" ,1): gp_settings.settings['TEXTHALIGN'] = "Right"
          else                                       : gp_error(errorstr)
      return

    if ((command == "unset") and autocomplete(linelist[1],"texthalign",5)):  # unset texthalign
     if (len(linelist) > 2):
      gp_error('Trailing word after "texthalign" not expected.')
     gp_settings.settings['TEXTHALIGN'] = gp_settings.settings_default['TEXTHALIGN']
     return

    if ((command == "set") and autocomplete(linelist[1],"textvalign",5)): # set textvalign
      test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
      if (test == None):
        gp_error("Error reading text vertical alignment.")
      else:
        errorstr = "Unrecognised vrtical alignment '%s'\nOptions are: 'top', 'centre' or 'bottom'.\n"%test.group(1)
        if (len(test.group(1).split()) > 1): gp_error(errorstr)
        else:
          if   autocomplete(test.group(1),"top"   ,1): gp_settings.settings['TEXTVALIGN'] = "Top"
          elif autocomplete(test.group(1),"centre",1): gp_settings.settings['TEXTVALIGN'] = "Centre"
          elif autocomplete(test.group(1),"center",1): gp_settings.settings['TEXTVALIGN'] = "Centre"
          elif autocomplete(test.group(1),"bottom",1): gp_settings.settings['TEXTVALIGN'] = "Bottom"
          else                                       : gp_error(errorstr)
      return

    if ((command == "unset") and autocomplete(linelist[1],"texthalign",5)):  # unset textvalign
     if (len(linelist) > 2):
      gp_error('Trailing word after "textvalign" not expected.')
     gp_settings.settings['TEXTVALIGN'] = gp_settings.settings_default['TEXTVALIGN']
     return

    if ((command == "set") and autocomplete(linelist[1],"axescolour",5)): # set axescolour
      test = re.match(r"""^\S*\s*\S*\s*(\S*)""",line)
      if (test == None):
        gp_error("Error reading colour.")
      else:
        test2 = re.match(r"""^\d\d*$""", test.group(1))
        if (test2 != None):
         gp_settings.settings['AXESCOLOUR'] = gp_settings.colour_list[(int(test.group(1))-1)%len(gp_settings.colour_list)]
        else:
         if (test.group(1).capitalize() in gp_settings.colours): gp_settings.settings['AXESCOLOUR'] = test.group(1).capitalize()
         else                                                  : gp_error("Unrecognised colour '%s'."%test.group(1))
      return

    if ((command == "unset") and autocomplete(linelist[1],"axescolour",5)):  # unset axescolour
     if (len(linelist) > 2):
      gp_error('Trailing word after "axescolour" not expected.')
     gp_settings.settings['AXESCOLOUR'] = gp_settings.settings_default['AXESCOLOUR']
     return

    if ((command == "set") and autocomplete(linelist[1],"gridmajcolour",8)): # set gridmajcolour
      test = re.match(r"""^\S*\s*\S*\s*(\S*)""",line)
      if (test == None):
        gp_error("Error reading colour.")
      else:
        test2 = re.match(r"""^\d\d*$""", test.group(1))
        if (test2 != None):
         gp_settings.settings['GRIDMAJCOLOUR'] = gp_settings.colour_list[(int(test.group(1))-1)%len(gp_settings.colour_list)]
        else:
         if (test.group(1).capitalize() in gp_settings.colours): gp_settings.settings['GRIDMAJCOLOUR'] = test.group(1).capitalize()
         else                                                  : gp_error("Unrecognised colour '%s'."%test.group(1))
      return

    if ((command == "unset") and autocomplete(linelist[1],"gridmajcolour",8)):  # unset gridmajcolour
     if (len(linelist) > 2):
      gp_error('Trailing word after "gridmajcolour" not expected.')
     gp_settings.settings['GRIDMAJCOLOUR'] = gp_settings.settings_default['GRIDMAJCOLOUR']
     return

    if ((command == "set") and autocomplete(linelist[1],"gridmincolour",8)): # set gridmincolour
      test = re.match(r"""^\S*\s*\S*\s*(\S*)""",line)
      if (test == None):
        gp_error("Error reading colour.")
      else:
        test2 = re.match(r"""^\d\d*$""", test.group(1))
        if (test2 != None):
         gp_settings.settings['GRIDMINCOLOUR'] = gp_settings.colour_list[(int(test.group(1))-1)%len(gp_settings.colour_list)]
        else:
         if (test.group(1).capitalize() in gp_settings.colours): gp_settings.settings['GRIDMINCOLOUR'] = test.group(1).capitalize()
         else                                                  : gp_error("Unrecognised colour '%s'."%test.group(1))
      return 

    if ((command == "unset") and autocomplete(linelist[1],"gridmincolour",8)):  # unset gridmincolour
     if (len(linelist) > 2):
      gp_error('Trailing word after "gridmincolour" not expected.')
     gp_settings.settings['GRIDMAJCOLOUR'] = gp_settings.settings_default['GRIDMINCOLOUR']
     return

    if ((command == "set") and autocomplete(linelist[1],"dpi",3)): # set dpi
      test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
      if (test == None):
        gp_error("Error reading number of dots per inch.")
      else:
        try:
          value = float(test.group(1))
          if (value < 1.0): raise ValueError # Negative DPIs are stupid
          gp_settings.settings['DPI'] = value
        except KeyboardInterrupt: raise
        except:
          gp_error("Could not read number of dots per inch. Should be a floating point value.")
      return

    if ((command == "unset") and autocomplete(linelist[1],"dpi",3)):  # unset dpi
     if (len(linelist) > 2):
      gp_error('Trailing word after "dpi" not expected.')
     gp_settings.settings['DPI'] = gp_settings.settings_default['DPI']
     return

    if ((command == "set") and (autocomplete(linelist[1],"pointsize",1) or autocomplete(linelist[1],"ps",2))): # set pointsize
      test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
      if (test == None):
        gp_error("Error reading pointsize")
      else:
        try:
          gp_settings.settings['POINTSIZE']=float(test.group(1))
        except KeyboardInterrupt: raise
        except:
          gp_error("Could not reading pointsize. Should be a floating point value")
      return

    if ((command == "unset") and (autocomplete(linelist[1],"pointsize",1) or autocomplete(linelist[1],"ps",2))):  # unset pointsize
     if (len(linelist) > 2):
      gp_error('Trailing word after "pointsize" not expected.')
     gp_settings.settings['POINTSIZE'] = gp_settings.settings_default['POINTSIZE']
     return

    if ((command == "set") and (autocomplete(linelist[1],"pointlinewidth",6) or autocomplete(linelist[1],"plw",3))): # set pointlinewidth EXTENDED API!!
      test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
      if (test == None):
        gp_error("Error reading pointlinewidth")
      else:
        try:
          gp_settings.settings['POINTLINEWIDTH']=float(test.group(1))
        except KeyboardInterrupt: raise
        except:
          gp_error("Could not reading pointlinewidth. Should be a floating point value")
      return

    if ((command == "unset") and (autocomplete(linelist[1],"pointlinewidth",6) or autocomplete(linelist[1],"plw",3))): # unset pointlinewidth EXTENDED API!!
     if (len(linelist) > 2):
      gp_error('Trailing word after "pointlinewidth" not expected.')
     gp_settings.settings['POINTLINEWIDTH'] = gp_settings.settings_default['POINTLINEWIDTH']
     return

    if ((command == "set") and (autocomplete(linelist[1],"linewidth",5) or autocomplete(linelist[1],"lw",2))): # set linewidth
      test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
      if (test == None):
        gp_error("Error reading linewidth")
      else:
        try:
          gp_settings.settings['LINEWIDTH']=float(test.group(1))
        except KeyboardInterrupt: raise
        except:
          gp_error("Could not reading linewidth. Should be a floating point value")
      return

    if ((command == "unset") and (autocomplete(linelist[1],"linewidth",5) or autocomplete(linelist[1],"lw",2))): # unset linewidth
     if (len(linelist) > 2):
      gp_error('Trailing word after "linewidth" not expected.')
     gp_settings.settings['LINEWIDTH'] = gp_settings.settings_default['LINEWIDTH']
     return

    if ((command == "set") and autocomplete(linelist[1],"linestyle",1)): # set linestyle
      test = re.match(r"""^\S*\s*\S*\s*([0-9][0-9]*)\s*(.*)""",line)
      if (test == None):
        gp_error("Error reading linestyle. Use format: set linestyle <tag> <linestyle>")
      else:
        try:
          ls_number = int(test.group(1))
        except KeyboardInterrupt: raise
        except:
          gp_error("linestyle number must be specified as an integer")
          return
        try:
          gp_settings.linestyles[ls_number] = test.group(2)
        except KeyboardInterrupt: raise
        except:
          gp_error("Unexpected error whilst writing linestyle %d"%ls_number)
      return

    if (((command == "set") and autocomplete(linelist[1],"nolinestyle",3)) or # set nolinestyle
        ((command == "unset") and autocomplete(linelist[1],"linestyle",1))  ): # unset linestyle
      test = re.match(r"""^\S*\s*\S*\s*([0-9][0-9]*)""",line)
      if (test == None):
        gp_error("Error reading nolinestyle. Use format: 'set nolinestyle <tag>' or 'unset linestyle <tag>'.")
      else:
        try:
          ls_number = int(test.group(1))
        except KeyboardInterrupt: raise
        except:
          gp_error("linestyle number must be specified as an integer.")
          return
        try:
          del gp_settings.linestyles[ls_number] # Delete key from linestyle dictionary
        except KeyboardInterrupt: raise
        except:
          gp_error("Error removing linestyle %d -- no such linestyle."%ls_number)
      return

    if ((command == "set") and autocomplete(linelist[1],"arrow",1)): # set arrow
      errorstring = "Error reading arrow. Use format: set arrow <tag> from (first|second|screen|graph) x, (first|second|screen|graph) y to (first|second|screen|graph) x, (first|second|screen|graph) y (with linestyle.... head|nohead|twoway)"
      try:
        test = re.match(r"^\S*\s\s*\S*\s\s*([0-9][0-9]*)\s\s*f(r(om?)?)?\s\s*(.*)""",line)
        if (test == None): raise SyntaxError
        arrow_number = int(test.group(1))
        restofline   = test.group(4).split(",")
        if (len(restofline) != 3): raise SyntaxError
        test = re.match(r"(.*) to? (.*)", restofline[1])
        test2= re.match(r"(.*) w(i(th?)?)? (.*)", restofline[2])
        if (test == None): raise SyntaxError
        restofline = [ restofline[0], test.group(1), test.group(2), restofline[2] ]
        if (test2 != None):
          restofline[3] = test2.group(1)
          withclause = test2.group(4).strip()
        else:
          withclause = ""
        coordsys = [None, None, None, None]
        x        = [None, None, None, None]
        for i in range(4):
         split       = restofline[i].strip().split(" ",1)
         coord       = read_coord_syst(split[0])
         coordsys[i] = coord[0]
         if (len(split) == 2):
          ordinate = coord[1] + " " + split[1]
         else:
          ordinate = coord[1]
         x[i]     = gp_eval.gp_eval(ordinate, gp_settings.variables, gp_settings.functions, verbose=False)
        if (len(withclause) != 0): arrow_style = withclause.split()
        else                     : arrow_style = []
        gp_settings.arrows[arrow_number] = (coordsys[0], x[0], coordsys[1], x[1], coordsys[2], x[2], coordsys[3], x[3], arrow_style)
      except KeyboardInterrupt: raise
      except:
       gp_error(errorstring)
       gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")
      return

    if (((command == "set") and autocomplete(linelist[1],"noarrow",2)) or # set noarrow
        ((command == "unset") and autocomplete(linelist[1],"arrow",1))  ): # unset arrow
      test = re.match(r"""^\S*\s*\S*\s*([0-9]*)""",line)
      if (test == None):
        gp_error("Error reading noarrow. Use format: 'set noarrow <tag>' or 'unset arrow <tag>'.")
      else:
        if (len(test.group(1)) == 0):
          gp_settings.arrows = {}
        else:
          try:
            arrow_number = int(test.group(1))
          except KeyboardInterrupt: raise
          except:
            gp_error("arrow number must be specified as an integer.")
            return
          try:
            del gp_settings.arrows[arrow_number] # Delete key from linestyle dictionary
          except KeyboardInterrupt: raise
          except:
            gp_error("Error removing arrow %d -- no such arrow."%arrow_number)
      return

    if (((command == "set"  ) and autocomplete(linelist[1],"grid",1  )) or # set grid
        ((command == "unset") and autocomplete(linelist[1],"nogrid",3)) ): # unset nogrid
      if (gp_settings.settings['GRID'] != 'ON'):
       gp_settings.settings['GRID']='ON'
       if (len(linelist) == 2):
        gp_settings.settings['GRIDAXISX']=gp_settings.settings_default['GRIDAXISX'][:] # set grid alone puts grid onto default axes
        gp_settings.settings['GRIDAXISY']=gp_settings.settings_default['GRIDAXISY'][:]
       else:
        gp_settings.settings['GRIDAXISX']=[] # set grid <axis> puts a grid only onto specified axis
        gp_settings.settings['GRIDAXISY']=[]
      try:
       test = re.match(r"""\s*\S*\s*\S*\s*(.*)""", line)
       lineleft = test.group(1)
       while ((test != None) and (len(lineleft.strip()) > 0)):
        test = re.match(r"""\s*(x|y)(\d*)(.*)""", lineleft)
        if (test != None):
         if (len(test.group(2)) == 0): axis_n = 1
         else                        : axis_n = int(test.group(2))
         if (axis_n < 1):
          gp_warning("Warning: attempt in set grid command to put a grid on axis number < 1.")
         else:
          if (test.group(1) == 'x'): # Put a grid on an x-axis
           if (axis_n not in gp_settings.settings['GRIDAXISX']):
            gp_settings.settings['GRIDAXISX'].append(axis_n)
           else:
            gp_warning("Warning: attempt to set grid on axis x%d; grid is already set."%axis_n)
          else: # Put a grid on a y-axis
           if (axis_n not in gp_settings.settings['GRIDAXISY']):
            gp_settings.settings['GRIDAXISY'].append(axis_n)
           else:
            gp_warning("Warning: attempt to set grid on axis y%d; grid is already set."%axis_n)
         lineleft = test.group(3)
       if (len(lineleft.strip()) > 0):
        gp_error("Error: grid axes declaration must take the form x<n>y<m> where {n,m}>0.")
        gp_error("       left over string '%s' could not be parsed."%lineleft)
      except KeyboardInterrupt: raise
      except:
       gp_error("Error: internal error in set grid statement.")

      return

    if (((command == "set") and autocomplete(linelist[1],"nogrid",3)) or # set nogrid
        ((command == "unset") and autocomplete(linelist[1],"grid",1)) ): # unset grid
      if (len(linelist) == 2):
       gp_settings.settings['GRID']='OFF'
      else:
       try:
        test = re.match(r"""\s*\S*\s*\S*\s*(.*)""", line)
        lineleft = test.group(1)
        while ((test != None) and (len(lineleft.strip()) > 0)):
         test = re.match(r"""\s*(x|y)(\d*)(.*)""", lineleft)
         if (test != None):
          if (len(test.group(2)) == 0): axis_n = 1
          else                        : axis_n = int(test.group(2))
          if (axis_n < 1):
           gp_warning("Warning: attempt in set nogrid command to put a grid on axis number < 1.")
          else:
           if (test.group(1) == 'x'): # Remove a grid from an x-axis
            if (axis_n in gp_settings.settings['GRIDAXISX']):
             gp_settings.settings['GRIDAXISX'].remove(axis_n)
            else:
             gp_warning("Warning: attempt to set nogrid on axis x%d; no grid on that axis."%axis_n)
           else: # Remove a grid from a y-axis
            if (axis_n in gp_settings.settings['GRIDAXISY']):
             gp_settings.settings['GRIDAXISY'].remove(axis_n)
            else:
             gp_warning("Warning: attempt to set nogrid on axis y%d; no grid on that axis."%axis_n)
          lineleft = test.group(3)
        if (len(lineleft.strip()) > 0):
         gp_error("Error: nogrid axes declaration must take the form x<n>y<m> where {n,m}>0.")
         gp_error("       left over string '%s' could not be parsed."%lineleft)
       except KeyboardInterrupt: raise
       except:
        gp_error("Error: internal error in set nogrid statement.")
        gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")
      return

    if autocomplete(linelist[1],"axis",2): # set axis | unset axis
      if (len(linelist) == 2):
       gp_settings.axes = {'x':{1:gp_settings.default_axis.copy()},
                           'y':{1:gp_settings.default_axis.copy()},
                           'z':{1:gp_settings.default_axis.copy()} }
      else:
       try:
        test = re.match(r"""\s*\S*\s*\S*\s*(.*)""", line)
        lineleft = test.group(1)
        while ((test != None) and (len(lineleft.strip()) > 0)):
         test = re.match(r"""\s*(x|y)(\d*)(.*)""", lineleft)
         if (test != None):
          if (len(test.group(2)) == 0): axis_n = 1
          else                        : axis_n = int(test.group(2))
          if (axis_n < 1):
           if (command == "unset"): gp_warning("Warning: attempt in unset axis command to configure axis number < 1.")
           else:                    gp_error  ("Error: attempt to set axis with number < 1.")
          else:
           direction = test.group(1)
           if (command == "unset"):
            if (axis_n in gp_settings.axes[direction]):
             del gp_settings.axes[direction][axis_n]
             if (axis_n in gp_plot.axes_this[direction]): del gp_plot.axes_this[direction][axis_n]
             if (axis_n == 1):
              gp_settings.axes[direction][1]  = gp_settings.default_axis.copy()
              gp_plot.axes_this[direction][1] = gp_settings.default_axis.copy()
            else:
             gp_warning("Warning: attempt to unset axis %s%d; no such axis."%(direction,axis_n))
           else:
            if (not axis_n in gp_settings.axes[direction]):
             gp_settings.axes[direction][axis_n] = gp_settings.default_axis.copy()
             gp_plot.axes_this[direction][axis_n] = {'SETTINGS':gp_settings.axes[direction][axis_n].copy(), 'MIN_USED':None, 'MAX_USED':None, 'AXIS':None}
          lineleft = test.group(3)
        if (len(lineleft.strip()) > 0):
         gp_error("Error: unset axis declaration must take the form x<n>y<m> where {n,m}>0.")
         gp_error("       left over string '%s' could not be parsed."%lineleft)
       except KeyboardInterrupt: raise
       except:
        gp_error("Error: internal error in unset axis statement.")
        gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")
      return

    if ((command == "set") and autocomplete(linelist[1],"label",2)):  # set label
      errorstring = "Error reading label. Use format: set label <tag> 'text' at (first|second|screen|graph) x, (first|second|screen|graph) y"
      try:
        test = re.match(r"""\S*\s\s*\S*\s\s*([0-9][0-9]*)\s*(('|").*)""",line)
        if (test == None):
          gp_error(errorstring) ; return
        label_number = int(test.group(1))
        [text, aftertext] = gp_eval.gp_getquotedstring(test.group(2))
        test = re.match(r"""\s\s*a?t?\s*(.*)""",aftertext)
        restofline   = test.group(1).split(",")
        coordsys = [None, None]
        x        = [None, None]
        if (len(restofline) != 2):
         gp_error(errorstring)
        else:
         for i in range(2):
          split       = restofline[i].strip().split(" ",1)
          coord       = read_coord_syst(split[0])
          coordsys[i] = coord[0]
          if (len(split) == 2):
           ordinate = coord[1] + " " + split[1]
          else:
           ordinate = coord[1]
          x[i]     = gp_eval.gp_eval(ordinate, gp_settings.variables, gp_settings.functions, verbose=False)
         gp_settings.labels[label_number] = (text, coordsys[0],x[0],coordsys[1],x[1])
      except KeyboardInterrupt: raise
      except:
        gp_error("Error reading label position:")
        gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")
      return

    if (((command == "set") and autocomplete(linelist[1],"nolabel",4)) or # set nolabel
        ((command == "unset") and autocomplete(linelist[1],"label",2))   ): # unset label
      test = re.match(r"""^\S*\s*\S*\s*([0-9]*)""",line)
      if (test == None):
        gp_error("Error reading nolabel. Use format: 'set nolabel <tag>' or 'unset label <tag>'.")
      else:
        if (len(test.group(1)) == 0):
          gp_settings.labels = {}
        else: 
          try:
            label_number = int(test.group(1))
          except KeyboardInterrupt: raise
          except:
            gp_error("label number must be specified as an integer.")
            return
          try:
            del gp_settings.labels[label_number] # Delete key from label dictionary
          except KeyboardInterrupt: raise
          except:
            gp_error("Error removing label %d -- no such label."%label_number)
      return

    if ((command == "set") and (len(linelist) > 2) and autocomplete(linelist[2],"style",1)):
      key=""
      if autocomplete(linelist[1],"data",1):
        key = "DATASTYLE" ; name = "data"
      if autocomplete(linelist[1],"function",1):
        key = "FUNCSTYLE" ; name = "function"
      if (len(key) > 0):
        test = re.match(r"""^\S*\s*\S*\s*\S*\s*(.*)""",line)
        if ((test == None) or (len(test.group(1)) == 0)):
          gp_error("Error reading %s style"%name)
          return
        else:
          if autocomplete(test.group(1),"lines",1)       : gp_settings.settings[key] = 'lines'       ; return
          if autocomplete(test.group(1),"points",1)      : gp_settings.settings[key] = 'points'      ; return
          if autocomplete(test.group(1),"lp",2)          : gp_settings.settings[key] = 'linespoints' ; return
          if autocomplete(test.group(1),"linespoints",5) : gp_settings.settings[key] = 'linespoints' ; return
          if autocomplete(test.group(1),"pl",2)          : gp_settings.settings[key] = 'linespoints' ; return
          if autocomplete(test.group(1),"pointslines",5) : gp_settings.settings[key] = 'linespoints' ; return
          if autocomplete(test.group(1),"dots",1)        : gp_settings.settings[key] = 'dots'        ; return
          if autocomplete(test.group(1),"boxes",1)       : gp_settings.settings[key] = 'boxes'       ; return
          if autocomplete(test.group(1),"wboxes",1)      : gp_settings.settings[key] = 'wboxes'      ; return
          if autocomplete(test.group(1),"impulses",1)    : gp_settings.settings[key] = 'impulses'    ; return
          if autocomplete(test.group(1),"steps",1)       : gp_settings.settings[key] = 'steps'       ; return
          if autocomplete(test.group(1),"fsteps",1)      : gp_settings.settings[key] = 'fsteps'      ; return
          if autocomplete(test.group(1),"histeps",1)     : gp_settings.settings[key] = 'histeps'     ; return
          if autocomplete(test.group(1),"errorbars",1)   : gp_settings.settings[key] = 'yerrorbars'  ; return
          if autocomplete(test.group(1),"xerrorbars",2)  : gp_settings.settings[key] = 'xerrorbars'  ; return
          if autocomplete(test.group(1),"yerrorbars",2)  : gp_settings.settings[key] = 'yerrorbars'  ; return
          if autocomplete(test.group(1),"xyerrorbars",2) : gp_settings.settings[key] = 'xyerrorbars' ; return
          if autocomplete(test.group(1),"errorrange",6)  : gp_settings.settings[key] = 'yerrorrange' ; return
          if autocomplete(test.group(1),"xerrorrange",7) : gp_settings.settings[key] = 'xerrorrange' ; return
          if autocomplete(test.group(1),"yerrorrange",7) : gp_settings.settings[key] = 'yerrorrange' ; return
          if autocomplete(test.group(1),"xyerrorrange",8): gp_settings.settings[key] = 'xyerrorrange'; return
          if autocomplete(test.group(1),"arrows_head",3)    : gp_settings.settings[key] = 'arrows_head'   ; return
          if autocomplete(test.group(1),"arrows_nohead",3)  : gp_settings.settings[key] = 'arrows_nohead' ; return
          if autocomplete(test.group(1),"arrows_twoway",3)  : gp_settings.settings[key] = 'arrows_twohead'; return
          if autocomplete(test.group(1),"arrows_twohead",3) : gp_settings.settings[key] = 'arrows_twohead'; return
          if autocomplete(test.group(1),"csplines",3)    : gp_settings.settings[key] = 'csplines'    ; return
          if autocomplete(test.group(1),"acsplines",3)   : gp_settings.settings[key] = 'acsplines'   ; return
          gp_error("Unknown %s style: '%s'"%(name,test.group(1)))
          return

    if ((command == "set") and autocomplete(linelist[1],"output",1)): # set output
      test = re.match(r"""^\S*\s*\S*\s*('|")(.*)('|")""",line)
      if (test == None):
        gp_error("Output filename should be placed in quotes.")
      else:
        gp_settings.settings['OUTPUT']=test.group(2)
      return

    if ((command == "unset") and autocomplete(linelist[1],"output",1)): # unset output
     if (len(linelist) > 2):
      gp_error('Trailing word after "output" not expected.')
     gp_settings.settings['OUTPUT'] = gp_settings.settings_default['OUTPUT']
     return

    if ((command == "set") and autocomplete(linelist[1],"backup",2)): # set backup
     if (len(linelist) > 2):
      gp_error('Trailing word after "set backup" not expected.')
     gp_settings.settings['BACKUP'] = "ON"
     return

    if ((command == "unset") and autocomplete(linelist[1],"backup",2) or # unset backup
       ((command == "set") and autocomplete(linelist[1],"nobackup",4))  ):
     if (len(linelist) > 2):
      gp_error('Trailing word after "backup" not expected.')
     gp_settings.settings['BACKUP'] = "OFF"
     return

    if ((command == "set") and autocomplete(linelist[1],"key",1)): # set key
      gp_settings.settings['KEY']='ON'
      if (len(linelist) > 2):
       gp_settings.settings['KEY_XOFF'] = gp_settings.settings['KEY_YOFF'] = 0.0 # Reset offset if position is specified
      keymode = 0 # First offset is x offset
      for i in range(2,len(linelist),1):
       if (autocomplete(linelist[i],"top"    ,1) or autocomplete(linelist[i],"bottom" ,1) or autocomplete(linelist[i],"ycentre",1) or autocomplete(linelist[i],"left"   ,1) or autocomplete(linelist[i],"right"  ,1) or autocomplete(linelist[i],"xcentre",1)):
        if (len(gp_settings.settings['KEYPOS'].split()) != 2): gp_settings.settings['KEYPOS'] = "TOP RIGHT" # Deal with if previously "below"
        if   autocomplete(linelist[i],"top"    ,1): gp_settings.settings['KEYPOS'] = 'TOP '    + gp_settings.settings['KEYPOS'].split()[1] ; continue
        elif autocomplete(linelist[i],"bottom" ,1): gp_settings.settings['KEYPOS'] = 'BOTTOM ' + gp_settings.settings['KEYPOS'].split()[1] ; continue
        elif autocomplete(linelist[i],"ycentre",1): gp_settings.settings['KEYPOS'] = 'MIDDLE ' + gp_settings.settings['KEYPOS'].split()[1] ; continue
        elif autocomplete(linelist[i],"left"   ,1): gp_settings.settings['KEYPOS'] = gp_settings.settings['KEYPOS'].split()[0] + ' LEFT'   ; continue
        elif autocomplete(linelist[i],"right"  ,1): gp_settings.settings['KEYPOS'] = gp_settings.settings['KEYPOS'].split()[0] + ' RIGHT'  ; continue
        elif autocomplete(linelist[i],"xcentre",1): gp_settings.settings['KEYPOS'] = gp_settings.settings['KEYPOS'].split()[0] + ' MIDDLE' ; continue
       if autocomplete(linelist[i],"below"  ,2): gp_settings.settings['KEYPOS'] = "BELOW"                                               ; continue
       if autocomplete(linelist[i],"outside",1): gp_settings.settings['KEYPOS'] = "OUTSIDE"                                             ; continue
       try:
        offset = gp_eval.gp_eval(linelist[i],gp_settings.variables,gp_settings.functions)
        if (keymode == 0): gp_settings.settings['KEY_XOFF'] = offset
        if (keymode == 1): gp_settings.settings['KEY_YOFF'] = offset
        if (keymode <  2): keymode = keymode + 1
        else             :
         gp_error("Too many key offset expressions specified! Specify one x and one y offset at most.")
         raise ValueError
        continue
       except KeyboardInterrupt: raise
       except: pass
      return

    if ((command == "set") and autocomplete(linelist[1],"nokey",3)): # set nokey
     if (len(linelist) > 2):
      gp_error('Trailing word after "nokey" not expected.')
     gp_settings.settings['KEY']='OFF'
     return

    if ((command == "unset") and autocomplete(linelist[1],"nokey",3)): # unset nokey
     if (len(linelist) > 2):
      gp_error('Trailing word after "nokey" not expected.')
     gp_settings.settings['KEY']='ON'
     return

    if ((command == "unset") and autocomplete(linelist[1],"key",1)): # unset key
     if (len(linelist) > 2):
      gp_error('Trailing word after "key" not expected.')
     gp_settings.settings['KEY']      = gp_settings.settings_default['KEY']
     gp_settings.settings['KEYPOS']   = gp_settings.settings_default['KEYPOS']
     gp_settings.settings['KEY_XOFF'] = gp_settings.settings_default['KEY_XOFF']
     gp_settings.settings['KEY_YOFF'] = gp_settings.settings_default['KEY_YOFF']
     return

    if ((command == "set") and autocomplete(linelist[1],"keycolumns",4)): # set keycolumns
      test = re.match(r"""^\S*\s*\S*\s*(.*)""",line)
      if (test == None):
        gp_error("Error reading number of columns for key.")
      else:
        try:
          keycols = int(test.group(1))
          assert keycols > 0
          gp_settings.settings['KEYCOLUMNS'] = keycols
        except KeyboardInterrupt: raise
        except:
          gp_error("Invalid number of columns for key. Should be an integer value > 0.")
      return

    if ((command == "unset") and autocomplete(linelist[1],"keycolumns",4)): # unset keycolumns
     if (len(linelist) > 2):
      gp_error('Trailing word after "keycolumns" not expected.')
     gp_settings.settings['KEYCOLUMNS'] = gp_settings.settings_default['KEYCOLUMNS']
     return

    if ((command == "set") and autocomplete(linelist[1],"terminal",1)): # set terminal
      for i in range(2,len(linelist),1):
        if autocomplete(linelist[i],"x11_singlewindow",1): gp_settings.settings['TERMTYPE'] = "X11_singlewindow" ; continue
        if autocomplete(linelist[i],"x11_multiwindow" ,1): gp_settings.settings['TERMTYPE'] = "X11_multiwindow"  ; continue
        if autocomplete(linelist[i],"postscript"      ,1): gp_settings.settings['TERMTYPE'] = "PS"               ; continue
        if autocomplete(linelist[i],"eps"             ,1): gp_settings.settings['TERMTYPE'] = "PS"               ; gp_settings.settings['ENHANCED'] = "ON" ; continue
        if autocomplete(linelist[i],"png"             ,3): gp_settings.settings['TERMTYPE'] = "PNG"              ; continue
        if autocomplete(linelist[i],"gif"             ,3): gp_settings.settings['TERMTYPE'] = "GIF"              ; continue
        if autocomplete(linelist[i],"jpg"             ,3): gp_settings.settings['TERMTYPE'] = "JPG"              ; continue
        if autocomplete(linelist[i],"jpeg"            ,3): gp_settings.settings['TERMTYPE'] = "JPG"              ; continue
        if autocomplete(linelist[i],"color"           ,1): gp_settings.settings['COLOUR']   = "ON"               ; continue
        if autocomplete(linelist[i],"colour"          ,1): gp_settings.settings['COLOUR']   = "ON"               ; continue
        if autocomplete(linelist[i],"monochrome"      ,1): gp_settings.settings['COLOUR']   = "OFF"              ; continue
        if autocomplete(linelist[i],"enhanced"        ,2): gp_settings.settings['ENHANCED'] = "ON"               ; continue
        if autocomplete(linelist[i],"noenhanced"      ,1): gp_settings.settings['ENHANCED'] = "OFF"              ; continue
        if autocomplete(linelist[i],"landscape"       ,1): gp_settings.settings['LANDSCAPE']= "ON"               ; continue
        if autocomplete(linelist[i],"portrait"        ,3): gp_settings.settings['LANDSCAPE']= "OFF"              ; continue
        if autocomplete(linelist[i],"transparent"     ,1): gp_settings.settings['TERMTRANSPARENT']= "ON"         ; continue
        if autocomplete(linelist[i],"solid"           ,1): gp_settings.settings['TERMTRANSPARENT']= "OFF"        ; continue
        if autocomplete(linelist[i],"invert"          ,1): gp_settings.settings['TERMINVERT']= "ON"              ; continue
        if autocomplete(linelist[i],"noinvert"        ,3): gp_settings.settings['TERMINVERT']= "OFF"             ; continue
        gp_error("Unrecognised word: %s"%linelist[i])
      return

    if ((command == "unset") and autocomplete(linelist[1],"terminal",1)): # unset terminal
     if (len(linelist) > 2):
      gp_error('Trailing word after "terminal" not expected.')
     gp_settings.settings['TERMTYPE'] = gp_settings.settings_default['TERMTYPE']
     gp_settings.settings['COLOUR']   = gp_settings.settings_default['COLOUR']
     return

    test = re.match("""^\S*\s*(x|y|z)(\d*)""",line) # set label / unset label
    if (test != None):
     name = test.group(1) + test.group(2)
     if autocomplete(linelist[1],"%slabel"%name,3):
      try:
       direction = test.group(1)
       if (len(test.group(2)) == 0): number = 1
       else                        : number = int(test.group(2))
       if (number < 1): raise ValueError # Don't let user assign axis zero!
      except ValueError:
       gp_error("Illegal axis number in %s %s<n>label statement"%(command,test.group(1)))
       return
      if (command == "unset"):
       if (not number in gp_settings.axes[direction]):
        gp_error('Error: Attempt to unset label on non-existant axis %s%s.'%(direction,number))
        return
       else:
         if (len(linelist) > 2): gp_error('Trailing word after "label" not expected.')
         gp_settings.axes[direction][number]['LABEL']=""
      else:
       if (not number in gp_settings.axes[direction]): gp_settings.axes[direction][number] = gp_settings.default_axis.copy() # Create axis if it doesn't already exist
       test = re.match("""^\S*\s*\S*\s*(('|").*)""",line)
       if (test == None):
        gp_error("Label should be placed in quotes. Defaulting to blank title.")
        gp_settings.axes[direction][number]['LABEL']=""
       else:
        [title, aftertitle] = gp_eval.gp_getquotedstring(test.group(1))
        if (title != None):
          gp_settings.axes[direction][number]['LABEL'] = title
        else:
          gp_error("Error: Mismatched quotes in set %slabel command."%name)
        if (aftertitle != None) and (len(aftertitle.strip()) > 0):
          gp_error("Unexpected values given after set %slabel command."%name)
      if (not number in gp_plot.axes_this[direction]):                  # Modify axes_this is gp_plot, so that replot takes account of changes
       gp_plot.axes_this[direction][number] = {'SETTINGS':gp_settings.axes[direction][number].copy(), 'MIN_USED':None, 'MAX_USED':None, 'AXIS':None}
      else:
       gp_plot.axes_this[direction][number]['SETTINGS']['LABEL'] = gp_settings.axes[direction][number]['LABEL']
      return

    test = re.match("""^\S*\s*(x|y|z)(\d*)""",line) # set ticdir / unset ticdir
    if (test != None):
     name = test.group(1) + test.group(2)
     if autocomplete(linelist[1],"%sticdir"%name,6):
      try:
       direction = test.group(1)
       if (len(test.group(2)) == 0): number = 1
       else                        : number = int(test.group(2))
       if (number < 1): raise ValueError # Don't let user assign axis zero!
      except ValueError:
       gp_error("Illegal axis number in %s %s<n>ticdir statement"%(command,test.group(1)))
       return
      if (command == "unset"):
       if (not number in gp_settings.axes[direction]):
        gp_error('Error: Attempt to unset ticdir on non-existant axis %s%s.'%(direction,number))
        return
       else:
        if (len(linelist) > 2): gp_error('Trailing word after "ticdir" not expected.')
        gp_settings.axes[direction][number]['TICDIR']=gp_settings.default_axis['TICDIR']
      else:
       if (not number in gp_settings.axes[direction]): gp_settings.axes[direction][number] = gp_settings.default_axis.copy() # Create axis if it doesn't already exist
       if (len(linelist) < 3):
        gp_error("Error: Need to specify tic direction after 'set ticdir' statement.")
        return
       else:
        if   autocomplete(linelist[2],"inward",1) : gp_settings.axes[direction][number]['TICDIR'] = "INWARD"
        elif autocomplete(linelist[2],"outward",1): gp_settings.axes[direction][number]['TICDIR'] = "OUTWARD"
        elif autocomplete(linelist[2],"both",1)   : gp_settings.axes[direction][number]['TICDIR'] = "BOTH"
        else: gp_error("Error: Unrecognised tic direction '%s'. Should be 'inward', 'outward' or 'both'."%linelist[2])
        if (len(linelist) > 3):
          gp_error('Error: Trailing word after "set ticdir" not expected.')
      if (not number in gp_plot.axes_this[direction]):                  # Modify axes_this is gp_plot, so that replot takes account of changes
       gp_plot.axes_this[direction][number] = {'SETTINGS':gp_settings.axes[direction][number].copy(), 'MIN_USED':None, 'MAX_USED':None, 'AXIS':None}
      else:
       gp_plot.axes_this[direction][number]['SETTINGS']['TICDIR'] = gp_settings.axes[direction][number]['TICDIR']
      return

    if ((autocomplete(linelist[1],"autoscale",2)) or (autocomplete(linelist[1],"logscale",2)) or (autocomplete(linelist[1],"nologscale",3))): # set log or nolog or autoscale
     if                             autocomplete(linelist[1],"autoscale",2)   : key = "AUTOSCALE" ; keyset = "ON"
     elif ((command == "set")   and autocomplete(linelist[1],"logscale",2)   ): key = "LOG"       ; keyset = "ON"
     elif ((command == "unset") and autocomplete(linelist[1],"logscale",2)   ): key = "LOG"       ; keyset = "OFF"
     elif ((command == "set")   and autocomplete(linelist[1],"nologscale",3) ): key = "LOG"       ; keyset = "OFF"
     elif ((command == "unset") and autocomplete(linelist[1],"nologscale",3) ): key = "LOG"       ; keyset = "ON"
     else:
      gp_error("Error: setting autoscale/logscale/nologscale, and getting very confused in the process.")
      return
     if (len(linelist) == 2): # set log on it own applies log scaling to all axes.
       axisnames = ""
       for direction in gp_settings.axes:
        for number in gp_settings.axes[direction]:
         axisnames += direction+str(number)
       linelist.append(axisnames)
     for i in range(2,len(linelist)):
      word = linelist[i]
      while (len(word) > 0):
       test = re.match("""(x|y|z)(\d*)(.*)""",word)
       if (test != None):
        try:
         direction = test.group(1)
         if (len(test.group(2)) == 0): number = 1
         else                        : number = int(test.group(2))
         word      = test.group(3)
         if (number < 1): raise ValueError # Don't let user assign axis zero!
        except ValueError:
         gp_error("Illegal axis name '%s%s' in set statement"%(direction,test.group(2)))
         return
        if (not number in gp_settings.axes[direction]):
          if (command == "set"):
            gp_settings.axes[direction][number] = gp_settings.default_axis.copy() # Create axis if it doesn't already exist
          else:
            gp_error('Error: Attempt to unset %s on non-existant axis %s%s.'%(linelist[1],direction,number))
            return
        if (key != "AUTOSCALE"):
         gp_settings.axes[direction][number][key] = keyset
        else:
         gp_settings.axes[direction][number]['MIN'] = gp_settings.axes[direction][number]['MAX'] = None
        if (not number in gp_plot.axes_this[direction]):                  # Modify axes_this is gp_plot, so that replot takes account of changes
         gp_plot.axes_this[direction][number] = {'SETTINGS':gp_settings.axes[direction][number].copy(), 'MIN_USED':None, 'MAX_USED':None, 'AXIS':None}
        else:
         if (key != "AUTOSCALE"):
          gp_plot.axes_this[direction][number]['SETTINGS'][key] = gp_settings.axes[direction][number][key]
         else:
          gp_settings.axes[direction][number]['MIN'] = gp_settings.axes[direction][number]['MAX'] = None
       else:
        gp_error("No such axis as %s"%word)
        break
     return

    test = re.match("""^\S*\s*(x|y|z)(\d*)""",line) # set range / unset range
    if (test != None):
     name = test.group(1) + test.group(2)
     if autocomplete(linelist[1],"%srange"%name,3):
      try:
       direction = test.group(1)
       if (len(test.group(2)) == 0): number = 1
       else:                         number = int(test.group(2))
       if (number < 1): raise ValueError # Don't let user assign axis zero!
      except ValueError:
       gp_error("Illegal axis number is set %s<n>range statement"%test.group(1))
       return
      if (command == "unset"):
       if (not number in gp_settings.axes[direction]):
        gp_error('Error: Attempt to unset range on non-existant axis %s%s.'%(direction,number))
        return
       else:
        if (len(linelist) > 2): gp_error('Trailing word after "range" not expected.')
        gp_settings.axes[direction][number]['MIN']       = gp_settings.default_axis['MIN']
        gp_settings.axes[direction][number]['MAX']       = gp_settings.default_axis['MAX']
      else:
       if (not number in gp_settings.axes[direction]): gp_settings.axes[direction][number] = gp_settings.default_axis.copy() # Create axis if it doesn't already exist
       test = re.match("""^\S*\s*\S*\s*restore""",line)
       if (test != None):
        gp_settings.axes[direction][number]['MIN'] = gp_settings.axes[direction][number]['MAX'] = None
       else:
        test = re.match(r"""^\S*\s*\S*\s*\[([^:\]]*)((:)|( *to *))?([^:\]]*)\]""",line)
        if (test == None):
         gp_error("Could not read range. Use format: set xrange [0:1].")
        else:
         try:
          if (len(test.group(1).strip()) != 0):
            gp_settings.axes[direction][number]['MIN'] = gp_eval.gp_eval(test.group(1),gp_settings.variables,gp_settings.functions)
          if (len(test.group(5).strip()) != 0):
            gp_settings.axes[direction][number]['MAX'] = gp_eval.gp_eval(test.group(5),gp_settings.variables,gp_settings.functions)
         except KeyboardInterrupt: raise
         except:
          gp_error("Error reading specified range.")
          gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")

      if (not number in gp_plot.axes_this[direction]):                  # Modify axes_this is gp_plot, so that replot takes account of changes
       gp_plot.axes_this[direction][number] = {'SETTINGS':gp_settings.axes[direction][number].copy(), 'MIN_USED':None, 'MAX_USED':None, 'AXIS':None}
      else:
       gp_plot.axes_this[direction][number]['SETTINGS']['MIN']       = gp_settings.axes[direction][number]['MIN']
       gp_plot.axes_this[direction][number]['SETTINGS']['MAX']       = gp_settings.axes[direction][number]['MAX']
      return

    if (command == "set"):
     gp_error(gp_text.set)
    else:
     gp_error(gp_text.unset)

# Main Directive Processor

line_combiner = ""

def directive(line, toplevel=True, interactive=False):
  global exitting, line_combiner

  if toplevel: gp_settings.cmd_history.append(line)

  gp_children.sigchld_handle()
  if (line.strip() == ""): return(2) # Blank lines do nothing.
  if (line.strip()[0] == "#"): return(2) # Comment lines also do nothing.

  # Check for \ line splitter
  if (line.strip()[-1] == "\\"):
   line_combiner = line_combiner + line.strip()[:-1]
   return
  line = line_combiner + line
  line_combiner = ""

  # Check for `` shell substitute command
  linesplit = gp_eval.gp_split(line, "`")
  linenew   = ""
  if ((len(linesplit)%2) == 0):
   gp_error("Error: mismatched ` in line.")
   return(1)
  for i in range(len(linesplit)):
   if ((i%2) == 1):
    os.chdir(gp_settings.cwd)
    shell_cmd = os.popen(linesplit[i],"r")
    linesplit[i] = shell_cmd.read().replace('\n',' ')
    shell_cmd.close()
    os.chdir(gp_settings.tempdir)
   linenew = linenew + linesplit[i]
  line = linenew

  # Can use ; to pass multiple commands on one line
  if (line.strip()[0] != "!"): # Don't split shell commandlines
   linelist = gp_eval.gp_split(line,';')
  else:
   linelist = [line.strip()]
  if (len(linelist) > 1):
   for i in range(len(linelist)):
    directive(linelist[i], False, interactive)
   return(2) 

  line     = line.strip() # Get rid of leading/trailing spaces
  linelist = line.split()
  if (len(linelist) < 1): return(2)

  # Can do plot[0:1] with no space between command and punctuation
  test = re.match("([A-Za-z][A-Za-z_]*)([^A-Za-z_].*)",linelist[0])
  if (test != None):
    linelist_new = [test.group(1), test.group(2)]
    linelist_new.extend(linelist[1:])
    linelist = linelist_new

  if   (re.match(r'([A-Za-z]\w*)(\([^()]*\))([^=]*)=(.*)',line.strip()) != None): # f(x) = ...
   t =  re.match(r'([A-Za-z]\w*)(\([^()]*\))([^=]*)=(.*)',line.strip())
   try:
    gp_eval.gp_function_declare(line.strip(), gp_settings.functions)
   except KeyboardInterrupt: raise
   except:
    gp_error("Error defining function %s:"%t.group(1))
    gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")
  elif (   re.match(r'^([A-Za-z]\w*)\s*=(.*)',line) != None):                     # x = ...
    test = re.match(r'^([A-Za-z]\w*)\s*=(.*)',line)
    try:
      if (len(test.group(2).strip()) == 0):
        del gp_settings.variables[test.group(1).strip()]
      else:
        gp_settings.variables[test.group(1).strip()] = gp_eval.gp_eval(test.group(2),gp_settings.variables,gp_settings.functions)
    except KeyboardInterrupt: raise
    except: pass

  elif (autocomplete(linelist[0],"exit",2) or      # exit / quit
        autocomplete(linelist[0],"quit",1)    ):
           exitting=1; return
  elif (line.strip()[0] == "!"):                   # the ! command
    shell_command = line.strip()[1:]
    os.chdir(gp_settings.cwd)
    os.system(shell_command) # Execute command in user's cwd
    os.chdir(gp_settings.tempdir)
  elif autocomplete(linelist[0],"cd",2):           # the cd command
    for i in range(1,len(linelist)):
     new_dir = glob.glob(os.path.join(gp_settings.cwd, linelist[i]))
     if (len(new_dir) == 0):
      gp_error("Error: Directory '%s' could not be found."%linelist[i])
     else:
      gp_settings.cwd = new_dir[0]
  elif autocomplete(linelist[0],"pwd",2):          # the pwd command
    gp_report(gp_settings.cwd)
  elif (autocomplete(linelist[0],"help",1) or      # help / ?
        autocomplete(linelist[0],"?",1)       ):
           gp_help.directive_help(line, interactive)
  elif autocomplete(linelist[0],"load",1):         # load
    test = re.match(r"""\s*\S*\s*('|")([^'"]*)('|")\s*$""",line)
    if (test == None):
      gp_error("""Use syntax "load 'filename'".""")
    else:
      main_loop([test.group(2)])
  elif autocomplete(linelist[0],"save",2):         # save
    test = re.match(r"""\s*\S*\s*('|")([^'"]*)('|")\s*$""",line)
    if (test == None):
      gp_error("""Use syntax "save 'filename'".""") 
    else:
      try:
       savefile = open(os.path.join(gp_settings.cwd, os.path.expanduser(test.group(2))),"w")
       for line in gp_settings.cmd_history[:-1]: savefile.write(line+"\n")
       savefile.close()
      except KeyboardInterrupt: raise
      except:
       gp_error("Error writing output to file '%s':"%test.group(2))
       gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")
  elif autocomplete(linelist[0],"set",2):          # set
    directive_set_unset("set",line,linelist)
  elif autocomplete(linelist[0],"unset",1):        # unset
    directive_set_unset("unset",line,linelist)
  elif autocomplete(linelist[0],"show",2):         # show
    directive_show(line,linelist)
  elif autocomplete(linelist[0],"fit",1):          # fit
    gp_fit.directive_fit(line,gp_settings.variables,gp_settings.functions)
  elif autocomplete(linelist[0],"spline",1):       # spline
    gp_spline.directive_spline(line,gp_settings.variables,gp_settings.functions)
  elif autocomplete(linelist[0],"plot",1):         # plot
    gp_plot.directive_plot(line,gp_settings.linestyles,gp_settings.variables,gp_settings.functions,gp_settings.settings,
                           gp_settings.axes,gp_settings.labels,gp_settings.arrows,0,interactive)
  elif autocomplete(linelist[0],"replot",3):       # replot
    gp_plot.directive_plot(line,gp_settings.linestyles,gp_settings.variables,gp_settings.functions,gp_settings.settings,
                           gp_settings.axes,gp_settings.labels,gp_settings.arrows,1,interactive)
  elif autocomplete(linelist[0],"text",3):         # text
    gp_plot.directive_text(line,gp_settings.linestyles,gp_settings.variables,gp_settings.functions,gp_settings.settings,interactive)
  elif autocomplete(linelist[0],"arrow",2):        # arrow
    gp_plot.directive_arrow(line,gp_settings.linestyles,gp_settings.variables,gp_settings.functions,gp_settings.settings,interactive)
  elif autocomplete(linelist[0],"clear",1):        # clear
    gp_plot.multiplot_plotdesc = [] ; gp_plot.multiplot_plotorder = [] ; gp_plot.multiplot_text = [] ; gp_plot.multiplot_arrows = [] ; gp_plot.multiplot_axes = []
    if (gp_children.ghostview_pid != None): os.kill(gp_children.ghostview_pid, signal.SIGKILL) # Kill any X11_singlewindow gv sessions
  elif autocomplete(linelist[0],"reset",3):        # reset
    gp_settings.settings              = gp_settings.settings_default.copy()
    gp_settings.settings['GRIDAXISX'] = gp_settings.settings_default['GRIDAXISX'][:]
    gp_settings.settings['GRIDAXISY'] = gp_settings.settings_default['GRIDAXISY'][:]
    gp_settings.axes = {'x':{1:gp_settings.default_axis.copy()},
                        'y':{1:gp_settings.default_axis.copy()},
                        'z':{1:gp_settings.default_axis.copy()} }

  elif autocomplete(linelist[0],"edit",1):       # edit
    if (gp_settings.settings['MULTIPLOT'] != 'ON'):
     gp_error("Error: Can only edit plots when in multiplot mode.")
    else:
     try:
      test = re.match(r"""^\s*%s\s*(.*)"""%linelist[0],line)
      editno = int(test.group(1))
      if (editno >= len(gp_plot.multiplot_plotdesc)) or (editno < 0):
       gp_error("Error: Attempt to edit a multiplot plot with index %d -- no such plot."%editno)
      else:
       gp_plot.replot_focus = editno
       gp_settings.settings = gp_plot.multiplot_plotdesc[editno][2].copy() # Reset all settings to those from this multiplot item
       gp_plot.plotlist     = gp_plot.multiplot_plotdesc[editno][0][:]
       gp_settings.labels   = gp_plot.multiplot_plotdesc[editno][3].copy()
       gp_settings.arrows   = gp_plot.multiplot_plotdesc[editno][4].copy()
       gp_settings.axes     = { 'x':{},'y':{},'z':{} } # Likewise for axis settings
       gp_plot.axes_this    = { 'x':{},'y':{},'z':{} }
       for [direction,axis_list_to] in gp_settings.axes.iteritems():
        for [number,axis] in gp_plot.multiplot_axes[editno][direction].iteritems():
         axis_list_to[number] = axis['SETTINGS'].copy()
         gp_plot.axes_this[direction][number] = {'SETTINGS':axis['SETTINGS'].copy(), 'MIN_USED':None, 'MAX_USED':None, 'AXIS':None}
     except KeyboardInterrupt: raise
     except:
      gp_error("Error: Please specify which item to edit after edit command")

  elif autocomplete(linelist[0],"delete",1):       # delete
    if (gp_settings.settings['MULTIPLOT'] != 'ON'):
     gp_error("Error: Can only delete a plot when in multiplot mode.")
    else:
     try:
      errorstr = "delete command should be followed by an integer indicating which plot to delete."
      test = re.match(r"""^\s*%s\s*(.*)"""%linelist[0],line)
      deleteno = int(test.group(1))
      if (deleteno >= len(gp_plot.multiplot_plotdesc)) or (deleteno < 0):
       gp_error("Error: Attempt to delete a multiplot plot with index %d -- no such plot."%deleteno)
      else:
       if (gp_plot.multiplot_plotdesc[deleteno][5] == 'ON'):
        gp_warning("Warning: Attempt to delete a multiplot plot which is already deleted.")
       else:
        gp_plot.multiplot_plotdesc[deleteno][5] = 'ON' # Set delete flag on this plot
        errorstr = "Problem encountered whilst refreshing display after delete operation."
        gp_plot.multiplot_plot(gp_settings.linestyles,gp_settings.variables,gp_settings.functions,gp_settings.settings) # Refresh display
     except KeyboardInterrupt: raise
     except:
      gp_error("Error: %s"%errorstr)
      gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")

  elif autocomplete(linelist[0],"undelete",3):       # undelete
    if (gp_settings.settings['MULTIPLOT'] != 'ON'):
     gp_error("Error: Can only undelete a plot when in multiplot mode.")
    else:
     try:
      errorstr = "undelete command should be followed by an integer indicating which plot to delete."
      test = re.match(r"""^\s*%s\s*(.*)"""%linelist[0],line)
      deleteno = int(test.group(1))
      if (deleteno >= len(gp_plot.multiplot_plotdesc)) or (deleteno < 0):
       gp_error("Error: Attempt to undelete a multiplot plot with index %d -- no such plot."%deleteno)
      else:
       if (gp_plot.multiplot_plotdesc[deleteno][5] != 'ON'):
        gp_warning("Warning: Attempt to undelete a multiplot plot which isn't deleted.")
       else:
        gp_plot.multiplot_plotdesc[deleteno][5] = 'OFF' # Unset delete flag on this plot
        errorstr = "Problem encountered whilst refreshing display after undelete operation."
        gp_plot.multiplot_plot(gp_settings.linestyles,gp_settings.variables,gp_settings.functions,gp_settings.settings) # Refresh display
     except KeyboardInterrupt: raise
     except:
      gp_error("Error: %s"%errorstr)
      gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")

  elif autocomplete(linelist[0],"move",1):           # move
    if (gp_settings.settings['MULTIPLOT'] != 'ON'):
     gp_error("Error: Can only move a plot when in multiplot mode.")
    else:
     try:
      errorstr = "Please specify which item to move after move command, and where to move it to. Syntax: 'move <n> x,y'."
      test = re.match(r"""^\s*%s\s*(\d\d*)\s*([^\s,]*)(\s*,?\s*)([^\s,]*)"""%linelist[0],line)
      moveno = int(test.group(1))
      xnew   = gp_eval.gp_eval(test.group(2),gp_settings.variables,gp_settings.functions)
      ynew   = gp_eval.gp_eval(test.group(4),gp_settings.variables,gp_settings.functions)
      if (len(test.group(3)) < 1): raise ValueError
      if (moveno >= len(gp_plot.multiplot_plotdesc)) or (moveno < 0):
       gp_error("Error: Attempt to move a multiplot plot with index %d -- no such plot."%moveno)
      else:
       gp_plot.multiplot_plotdesc[moveno][2]['ORIGINX'] = xnew
       gp_plot.multiplot_plotdesc[moveno][2]['ORIGINY'] = ynew
       errorstr = "Problem encountered whilst refreshing display after move  operation."
       gp_plot.multiplot_plot(gp_settings.linestyles,gp_settings.variables,gp_settings.functions,gp_settings.settings) # Refresh display
     except KeyboardInterrupt: raise
     except:
      gp_error("Error: %s"%errorstr)
      gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")

  elif autocomplete(linelist[0],"delete_text",8):  # delete_text
    if (gp_settings.settings['MULTIPLOT'] != 'ON'):
     gp_error("Error: Can only delete a text item when in multiplot mode.")
    else:
     try:
      errorstr = "delete_text command should be followed by an integer indicating which text item to delete."
      test = re.match(r"""^\s*%s\s*(.*)"""%linelist[0],line)
      deleteno = int(test.group(1))
      if (deleteno >= len(gp_plot.multiplot_text)) or (deleteno < 0):
       gp_error("Error: Attempt to delete a multiplot text item with index %d -- no such item."%deleteno)
      else:
       if (gp_plot.multiplot_text[deleteno][4] == 'ON'):
        gp_warning("Warning: Attempt to delete a multiplot text item which is already deleted.")
       else:
        gp_plot.multiplot_text[deleteno][4] = 'ON' # Set delete flag on this plot
        errorstr = "Problem encountered whilst refreshing display after delete_text operation."
        gp_plot.multiplot_plot(gp_settings.linestyles,gp_settings.variables,gp_settings.functions,gp_settings.settings) # Refresh display
     except KeyboardInterrupt: raise
     except:
      gp_error("Error: %s"%errorstr)
      gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")

  elif autocomplete(linelist[0],"undelete_text",10): # undelete_text
    if (gp_settings.settings['MULTIPLOT'] != 'ON'):
     gp_error("Error: Can only undelete a text item when in multiplot mode.")
    else:
     try:
      errorstr = "undelete_text command should be followed by an integer indicating which text item to undelete."
      test = re.match(r"""^\s*%s\s*(.*)"""%linelist[0],line)
      deleteno = int(test.group(1))
      if (deleteno >= len(gp_plot.multiplot_text)) or (deleteno < 0):
       gp_error("Error: Attempt to undelete a multiplot text item with index %d -- no such item."%deleteno)
      else:
       if (gp_plot.multiplot_text[deleteno][4] != 'ON'):
        gp_warning("Warning: Attempt to undelete a multiplot text item which isn't deleted.")
       else:
        gp_plot.multiplot_text[deleteno][4] = 'OFF' # Unset delete flag on this plot
        errorstr = "Problem encountered whilst refreshing display after undelete_text operation."
        gp_plot.multiplot_plot(gp_settings.linestyles,gp_settings.variables,gp_settings.functions,gp_settings.settings) # Refresh display
     except KeyboardInterrupt: raise
     except:
      gp_error("Error: %s"%errorstr)
      gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")

  elif autocomplete(linelist[0],"move_text",6):    # move_text
    if (gp_settings.settings['MULTIPLOT'] != 'ON'):
     gp_error("Error: Can only move text items when in multiplot mode.")
    else:
     try:
      errorstr = "Please specify which item to move after move_text command, and where to move it to. Syntax: 'move_text <n> x,y'."
      test = re.match(r"""^\s*%s\s*(\d\d*)\s*([^\s,]*)(\s*,?\s*)([^\s,]*)"""%linelist[0],line)
      moveno = int(test.group(1))
      xnew   = gp_eval.gp_eval(test.group(2),gp_settings.variables,gp_settings.functions)
      ynew   = gp_eval.gp_eval(test.group(4),gp_settings.variables,gp_settings.functions)
      if (len(test.group(3)) < 1): raise ValueError
      if (moveno >= len(gp_plot.multiplot_text)) or (moveno < 0):
       gp_error("Error: Attempt to move a multiplot text item with index %d -- no such item."%moveno)
      else:
       gp_plot.multiplot_text[moveno][1] = xnew
       gp_plot.multiplot_text[moveno][2] = ynew
       errorstr = "Problem encountered whilst refreshing display after move_text operation."
       gp_plot.multiplot_plot(gp_settings.linestyles,gp_settings.variables,gp_settings.functions,gp_settings.settings) # Refresh display
     except KeyboardInterrupt: raise
     except:
      gp_error("Error: %s"%errorstr)
      gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")

  elif autocomplete(linelist[0],"delete_arrow",8):  # delete_arrow
    if (gp_settings.settings['MULTIPLOT'] != 'ON'):
     gp_error("Error: Can only delete an arrow when in multiplot mode.")
    else:
     try:
      errorstr = "delete_arrow command should be followed by an integer indicating which arrow to delete."
      test = re.match(r"""^\s*%s\s*(.*)"""%linelist[0],line)
      deleteno = int(test.group(1))
      if (deleteno >= len(gp_plot.multiplot_arrows)) or (deleteno < 0):
       gp_error("Error: Attempt to delete a multiplot arrow with index %d -- no such arrow."%deleteno)
      else:
       if (gp_plot.multiplot_arrows[deleteno][6] == 'ON'):
        gp_warning("Warning: Attempt to delete a multiplot arrow which is already deleted.")
       else:
        gp_plot.multiplot_arrows[deleteno][6] = 'ON' # Set delete flag on this plot
        errorstr = "Problem encountered whilst refreshing display after delete_arrow operation."
        gp_plot.multiplot_plot(gp_settings.linestyles,gp_settings.variables,gp_settings.functions,gp_settings.settings) # Refresh display
     except KeyboardInterrupt: raise
     except:
      gp_error("Error: %s"%errorstr)
      gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")

  elif autocomplete(linelist[0],"undelete_arrow",10): # undelete_arrow
    if (gp_settings.settings['MULTIPLOT'] != 'ON'):
     gp_error("Error: Can only undelete an arrow when in multiplot mode.")
    else:
     try:
      errorstr = "undelete_arrow command should be followed by an integer indicating which arrow to undelete."
      test = re.match(r"""^\s*%s\s*(.*)"""%linelist[0],line)
      deleteno = int(test.group(1))
      if (deleteno >= len(gp_plot.multiplot_arrows)) or (deleteno < 0):
       gp_error("Error: Attempt to undelete a multiplot arrow with index %d -- no such arrow."%deleteno)
      else:
       if (gp_plot.multiplot_arrows[deleteno][6] != 'ON'):
        gp_warning("Warning: Attempt to undelete a multiplot arrow which isn't deleted.")
       else:
        gp_plot.multiplot_arrows[deleteno][6] = 'OFF' # Unset delete flag on this plot
        errorstr = "Problem encountered whilst refreshing display after undelete_arrow operation."
        gp_plot.multiplot_plot(gp_settings.linestyles,gp_settings.variables,gp_settings.functions,gp_settings.settings) # Refresh display
     except KeyboardInterrupt: raise
     except:
      gp_error("Error: %s"%errorstr)
      gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")

  elif autocomplete(linelist[0],"refresh",3):      # refresh
    gp_plot.multiplot_plot(gp_settings.linestyles,gp_settings.variables,gp_settings.functions,gp_settings.settings) # Refresh display

  elif autocomplete(linelist[0],"print",2):        # print
    printstring = ""
    test = re.match(r'\s*%s\s*(.*)'%linelist[0],line)
    if (test != None):
      args = gp_eval.gp_split(test.group(1),",")   # print , separated list
      for arg in args:
        test = re.match(r"""^\s*'(.*)'\s*$""",arg)
        if (test != None): sys.stdout.write(test.group(1)); continue
        test = re.match(r"""^\s*"(.*)"\s*$""",arg)
        if (test != None): sys.stdout.write(test.group(1)); continue
        try:
          printstring += floatprint( gp_eval.gp_eval(arg,gp_settings.variables,gp_settings.functions) )
        except KeyboardInterrupt: raise
        except: pass
        printstring += " "
    gp_report(printstring)

  else:
    gp_error(gp_text.invalid)                      # invalid cmd
    return(1)
  return (0)

def Interactive():   # Interactive PyXPlot terminal
  global exitting,line_combiner
  exitting=0
  if (sys.stdin.isatty() and gp_settings.display_splash): # Only print welcome blurb if running interactively, i.e. not from pipe
    gp_report(gp_text.init)
  try:
    while (exitting==0):
      if (sys.stdin.isatty()):
        try:
          if (line_combiner == ""): prompt = "pyxplot> "
          else                    : prompt = ".......> "
          input_cmd = raw_input(prompt)
          try: directive(input_cmd, interactive=True)
          except KeyboardInterrupt: gp_report("Received SIGINT. Terminating command.")
        except KeyboardInterrupt: gp_report("") # CTRL-C at command-prompt just gives a new command prompt
      else:
        directive(raw_input()) # Don't print command prompt when running from a pipe
  except (KeyboardInterrupt,EOFError): pass # EOFError == CTRL-D
  if sys.stdin.isatty():
   if gp_settings.display_splash: gp_report("\nGoodbye. Have a nice day.")
   else                         : gp_report("")

# Main loop

recurse_depth = 0

def main_loop(commandparams):
 global recurse_depth 
 recurse_depth = recurse_depth + 1 # Recursive loading protection
 if (recurse_depth > 10):
  gp_warning("Warning: recursive file loading detecting; load command failing")
  return

 if (len(commandparams) > 0): # Input files specified on commandline
  for i in range(0,len(commandparams)):
   if (commandparams[i] == '-'): # A minus on commandline means interactive
    Interactive()
   elif ((commandparams[i] == '-h') or (commandparams[i] == '--help')):
    gp_report(gp_text.help)
   elif ((commandparams[i] == '-v') or (commandparams[i] == '--version')): # NB: -q option implemented below
    gp_report(gp_text.version)
   else:
    infiles = glob.glob(os.path.join(gp_settings.cwd, os.path.expanduser(commandparams[i])))
    if (len(infiles) == 0):
     gp_error("PyXPlot Error: Could not find command file '%s'"%commandparams[i])
     gp_error("Skipping on to next command file")
    else:
     for infile in infiles:
      exitting=0
      try:
       instream = open(infile,"r")
      except KeyboardInterrupt: raise
      except:
       gp_error("PyXPlot Error: Could not open command file '%s'"%commandparams[i])
       gp_error("Skipping on to next command file")
      else:
       firstline=True
       for line in instream.readlines():
        status = directive(line)
        if (firstline and (status == 1)):
         gp_error("Error on first line of commandfile: Is this is valid script?")
         gp_error("Aborting")
         break
        if (status == 0): firstline=False
        if (exitting==1): break
 else: # Otherwise enter interactive mode
   Interactive()

# MAIN ENTRY POINT

# Store path to user's cwd ; but put LaTeX's junk in /tmp for tidiness
gp_settings.cwd = os.getcwd()
os.chdir(gp_settings.tempdir)

# Read user's PyXPlot history, if it exists
readline.set_history_length(1000)
try: readline.read_history_file(os.path.expanduser("~/.pyxplot_history"))
except: pass

# Turn off splashscreen if requested (commandline option -q)
while '-q' in sys.argv:
 gp_settings.display_splash = False
 sys.argv.remove('-q')
while '--quiet' in sys.argv:
 gp_settings.display_splash = False
 sys.argv.remove('--quiet')

# Turn on splashscreen if requested (commandline option -V)
while '-V' in sys.argv:
 gp_settings.display_splash = False
 sys.argv.remove('-V')
while '--verbose' in sys.argv:
 gp_settings.display_splash = False
 sys.argv.remove('--verbose')

# Turn on syntax highlighting if requested (commandline option -c)
while '-c' in sys.argv:
 gp_error_setcolour()
 sys.argv.remove('-c')
while '--colour' in sys.argv:
 gp_error_setcolour()
 sys.argv.remove('--colour')
while '--color' in sys.argv:
 gp_error_setcolour()
 sys.argv.remove('--color')

# Turn off syntax highlighting if requested (commandline option -m)
while '-m' in sys.argv:
 gp_error_setnocolour()
 sys.argv.remove('-m')
while '--monochrome' in sys.argv:
 gp_error_setnocolour()
 sys.argv.remove('--monochrome')

# Fix default terminal. If running interatively at any point, use X11_singlewindow.
# If we are going to run entirely from script all the way, default is postscript
if (gp_settings.config_lookup_opt('settings','TERMTYPE','default',gp_settings.termtypes ) == 'default'):
 if  (  ((len(sys.argv)>1) and ("-" not in sys.argv))
     or (gp_version.GHOSTVIEW == '/bin/false')
     or ('DISPLAY' not in os.environ.keys())
     or (len(os.environ['DISPLAY']) < 1)
     or (not sys.stdin.isatty())
     ):
  gp_settings.settings['TERMTYPE'] = 'PS'

# Loop over all config files passed to us on the commandline
main_loop(sys.argv[1:])

# Clean up any temporary .eps files which X11 terminal has placed in /tmp
gp_children.massacre_children()
os.chdir("/tmp")
os.system("rm -Rf %s"%gp_settings.tempdir)

# Write history file
try: readline.write_history_file(os.path.expanduser("~/.pyxplot_history"))
except: pass
